/*
 *  Copyright 2003, 2005 Adrian Thurston <thurston@cs.queensu.ca>
 */

/*  This file is part of Kelbt.
 *
 *  Kelbt is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  Kelbt is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with Kelbt; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

#include <iostream>
#include <stdlib.h>
#include <ctype.h>
#include <limits.h>
#include "kelbt.h"
#include "parsedata.h"
#include "avlmap.h"
#include "avlbasic.h"
#include "avlset.h"
#include "mergesort.h"

using std::cerr;
using std::endl;

#define FRESH_BLOCK 8128

void ParseData::writeReference( ostream &out, Definition *prod, char *data )
{
	bool user = false, refThis = false;
	int rhsPos = -1;
	
	if ( data[0] == '$' )
		user = true;
	
	if ( data[1] == '$' || data[1] == '@' )
		refThis = true;
	else
		rhsPos = atoi(data+1);
		
	out << "(";
	if ( user )
		out << "&";
	
	/* Write the access to the language element. */
	if ( refThis )
		out << "redLel";
	else
		out << "rhs[" << (rhsPos-1) << "]";

	/* Get the language element. */
	KlangEl *lel = refThis ? prod->prodName : prod->rhsLelVect[rhsPos-1];

	if ( user ) {
		if ( lel == 0 || lel->type == KlangEl::Term ) {
			/* Terminal (maybe a lit). */
			out << "->user.token";
		}
		else if ( lel->type == KlangEl::NonTerm && lel->typeDef != 0 )
			out << "->user." << lel->typeDef->data;
		else {
			cerr << "error: do not have a user type" << endl;
			exit(1);
		}
	}

	out << ")";
}


void ParseData::writeInlinList( ostream &out, Definition *prod, InlineList *inlineList )
{
	for ( InlineList::Iter item = *inlineList; item.lte(); item++ ) {
		switch ( item->type ) {
		case InlineItem::Text:
			out << item->data;
			break;
		case InlineItem::Reference:
			writeReference( out, prod, item->data );
			break;
		}
	}
	out << "\n";
}


void ParseData::writeTokenDefs( ostream &out )
{
	for ( LelList::Iter lel = langEls; lel.lte(); lel++ ) {
		if ( lel->isUserTerm || lel == eofKlangEl ) {
			/* Write out the token that will be generated by the scanner. */
			out << "#define " << lel->data << " " << lel->id << "\n";
		}
	}
	out << "\n";
}

void ParseData::writeTokenIds( ostream &out )
{
	out << "/*\n";
	for ( LelList::Iter lel = langEls; lel.lte(); lel++ ) {
		if ( lel->data != 0 )
			out << "	" << lel->data << " " << lel->id << endl;
		else
			out << "	" << lel->id << endl;
	}
	out << "*/\n\n";
}

/* Write the classes representing the non-terminal language elements. */
void ParseData::writeLangEls( ostream &out )
{
	for ( TypeDefMap::Iter td = typeDefMap; td.lte(); td++ ) {
		/* Start the class representing the non-terminal. */
		out <<
			"struct Lel_" << td->key << "\n"
			"{\n";

		out << "#line " << td->value->typeBlock->loc.line << " \"" << fileName << "\"\n";
		writeInlinList( out, 0, td->value->typeBlock->inlineList );
		out << "\n#line " << outFilter->line + 1 << " \"" << outputFile << "\"\n";

		/* Close the struct off. */
		out <<
			"};\n"
			"\n";
	}

	out << 
		"union UserData\n"
		"{\n";

	for ( TypeDefMap::Iter td = typeDefMap; td.lte(); td++ )
		out << "	Lel_" << td->key << " " << td->key << ";\n";

	out << 
		"	Token token;\n"
		"};\n"
		"\n";

	out <<
		"#define act_shift 1\n"
		"#define act_reduce 2\n"
		"#define act_sr 3\n"
		"#define act_sb 0x1\n"
		"#define act_rb 0x2\n"
		"#define lower 0x0000ffff\n"
		"#define upper 0xffff0000\n";

	out << 
		"struct LangEl\n"
		"{\n"
		"	char *file;\n"
		"	int line;\n"
		"	int type;\n"
		"	int reduction;\n" /* Should be able to extract this from the FSM. */
		"	int state;\n"
		"	UserData user;\n";

	if ( elementBlock != 0 ) {
		out << "#line " << elementBlock->loc.line << " \"" << fileName << "\"\n";
		writeInlinList( out, 0, elementBlock->inlineList );
		out << "\n#line " << outFilter->line + 1 << " \"" << outputFile << "\"\n";
	}

	out <<
		"	unsigned int retry;\n"
		"	LangEl *next, *child;\n"
		"};\n"
		"\n";

}


void ParseData::startCodeGen()
{
    /* Write the preprocessor line info for going to the output file. */
	outStream->flush();
	*outStream << "#line " << outFilter->line + 1 << 
			" \"" << outputFile << "\"\n";
}


void ParseData::endCodeGen( int endLine )
{
    /* Write the preprocessor line info for to the input file. */
    *outStream << "#line " << endLine << " \"" << fileName << "\"\n";
}


void ParseData::writeHeader( ostream &out )
{
	out << 
		"	LangEl *freshEl;\n"
		"	int freshPos;\n"
		"	LangEl *pool;\n"
		"	int numRetry;\n"
		"	int numNodes;\n"
		"	LangEl *stackTop;\n"
		"	LangEl *lastFinal;\n"
		"	int errCount;\n"
		"	int curs;\n"
		"\n";

	out << 
		"	int init();\n"
		"	int parseLangEl( int type, const Token &token );\n"
		"	int finish();\n"
		"\n";
}

void ParseData::writeReduceBlocks( ostream &out )
{
	out << "switch ( reduction ) {\n";
	for ( DefList::Iter prod = prodList; prod.lte(); prod++ ) {
		/* If there is reduction code, embed it here. */
		if ( prod->redBlock != 0 ) {
			out << "case " << prod->prodId << ": {\n";

			out << "#line " << prod->redBlock->loc.line << " \"" << fileName << "\"\n";
			writeInlinList( out, prod, prod->redBlock->inlineList );
			out << "\n#line " << outFilter->line + 1 << " \"" << outputFile << "\"\n";

			out << "} break;\n";
		}
	}
	out << "}\n";
}

void ParseData::writeUndoBlocks( ostream &out )
{
	out << "switch ( redLel->reduction ) {\n";
	for ( DefList::Iter prod = prodList; prod.lte(); prod++ ) {
		/* If there is reduction code, embed it here. */
		if ( prod->undoBlock != 0 ) {
			out << "case " << prod->prodId << ": {\n";

			out << "#line " << prod->undoBlock->loc.line << " \"" << fileName << "\"\n";
			writeInlinList( out, prod, prod->undoBlock->inlineList );
			out << "\n#line " << outFilter->line + 1 << " \"" << outputFile << "\"\n";

			out << "} break;\n";
		}
	}
	out << "}\n";
}

void ParseData::writeFinalBlocks( ostream &out )
{
	out << "switch ( lel->reduction ) {\n";
	for ( DefList::Iter prod = prodList; prod.lte(); prod++ ) {
		/* If there is reduction code, embed it here. */
		if ( prod->finalBlock != 0 ) {
			out << "case " << prod->prodId << ": {\n";

			out << "#line " << prod->finalBlock->loc.line << " \"" << fileName << "\"\n";
			writeInlinList( out, prod, prod->finalBlock->inlineList );
			out << "\n#line " << outFilter->line + 1 << " \"" << outputFile << "\"\n";

			out << "} break;\n";
		}
	}
	out << "}\n";
}

void ParseData::writeParser( ostream &out )
{
	out <<
		"extern "; UARRAY_TYPE(maxLelId) << " keys[];\n"
		"extern "; UARRAY_TYPE(maxState) << " targs[];\n"
		"extern "; SARRAY_TYPE(maxIndex) << " indicies[];\n"
		"extern unsigned int actInds[];\n"
		"extern unsigned int actions[];\n"
		"extern int commitLen[];\n"
		"extern "; UARRAY_TYPE(maxOffset) << " offsets[];\n"
		"extern unsigned int fssProdIdIndex[];\n"
		"extern "; SARRAY_TYPE(maxProdLen) << " fssProdLengths[];\n"
		"extern "; UARRAY_TYPE(maxLelId) << " prodLhsIds[];\n"
		"extern char *prodNames[];\n"
		"extern char *lelNames[];\n"
		"extern char prodCommit[];\n"
		"extern unsigned int startState;\n"
		"\n";

	out << 
		"int " << fsmName << "::init()\n"
		"{\n"
		"	curs = startState;\n"
		"	pool = 0;\n"
		"	freshEl = new LangEl[" << FRESH_BLOCK << "];\n"
		"	#ifdef LOG_ACTIONS\n"
		"	cerr << \"allocating " << FRESH_BLOCK << " LangEls\" << endl;\n"
		"	#endif\n"
		"	stackTop = freshEl;\n"
		"	stackTop->type = 0;\n"
		"	stackTop->state = -1;\n"
		"	freshPos = 1;\n"
		"	lastFinal = stackTop;\n"
		"	numRetry = 0;\n"
		"	numNodes = 0;\n"
		"	errCount = 0;\n"
		"	return 0;\n"
		"}\n"
		"\n";

	out <<
		"#define reject() induceReject = true\n"
		"\n";

	out << 
		"int " << fsmName << "::parseLangEl( int type, const Token &token )\n"
		"{\n"
		"	int pos, targState;\n"
		"	unsigned int *action;\n"
		"	int rhsLen;\n"
		"	LangEl *rhs[32];\n"
		"	bool induceReject;\n"
		"\n"
		"	if ( curs < 0 )\n"
		"		return 0;\n"
		"\n"
		"	LangEl *input;\n"
		"	if ( pool == 0 ) {\n"
		"		if ( freshPos == " << FRESH_BLOCK << " ) {\n"
		"			freshEl = new LangEl[" << FRESH_BLOCK << "];\n"
		"			#ifdef LOG_ACTIONS\n"
		"			cerr << \"allocating " << FRESH_BLOCK << " LangEls\" << endl;\n"
		"			#endif\n"
		"			freshPos = 0;\n"
		"		}\n"
		"		input = freshEl + freshPos++;\n"
		"	}\n"
		"	else {\n"
		"		input = pool;\n"
		"		pool = pool->next;\n"
		"	}\n"
		"	numNodes += 1;\n"
		"	input->type = type;\n"
		"	input->user.token = token;\n"
		"	input->next = 0;\n"
		"	input->retry = 0;\n"
		"	input->child = 0;\n"
		"\n"
		"again:\n"
		"	if ( input == 0 )\n"
		"		return 0;\n"
		"\n";

	if ( translateBlock != 0 ) {
		out << "#line " << translateBlock->loc.line << " \"" << fileName << "\"\n";
		writeInlinList( out, 0, translateBlock->inlineList );
		out << "\n#line " << outFilter->line + 1 << " \"" << outputFile << "\"\n";
	}

	out <<
		"	LangEl *lel = input;\n"
		"	if ( lel->type < keys[curs<<1] || lel->type > keys[(curs<<1)+1] )\n"
		"		goto parseError;\n"
		"\n"
		"	pos = indicies[offsets[curs] + (lel->type - keys[curs<<1])];\n"
		"	if ( pos < 0 )\n"
		"		goto parseError;\n"
		"\n"
		"	induceReject = false;\n"
		"	targState = targs[pos];\n"
		"	action = actions + actInds[pos];\n"
		"	if ( lel->retry & lower )\n"
		"		action += (lel->retry & lower);\n"
		"\n"
		"	if ( *action & act_sb ) {\n"
		"		#ifdef LOG_ACTIONS\n"
		"		cerr << \"shifted: \" << lelNames[lel->type];\n"
		"		#endif\n"
		"		input = input->next;\n"
		"		lel->state = curs;\n"
		"		lel->next = stackTop;\n"
		"		stackTop = lel;\n"
		"\n"
		"		if ( action[1] == 0 )\n"
		"			lel->retry &= upper;\n"
		"		else {\n"
		"			lel->retry += 1;\n"
		"			numRetry += 1;\n"
		"			#ifdef LOG_ACTIONS\n"
		"			cerr << \" retry: \" << stackTop;\n"
		"			#endif\n"
		"		}\n"
		"		#ifdef LOG_ACTIONS\n"
		"		cerr << endl;\n"
		"		#endif\n"
		"	}\n"
		"\n"
		"	if ( commitLen[pos] != 0 ) {\n"
		"		#ifdef LOG_ACTIONS\n"
		"		cerr << \"running commit of length: \" << commitLen[pos] << endl;\n"
		"		#endif\n"
		"\n"
		"		LangEl *commitHead = stackTop;\n"
		"		int absCommitLen = commitLen[pos];\n"
		"		if ( absCommitLen < 0 ) {\n"
		"			commitHead = commitHead->next;\n"
		"			absCommitLen = -1 * absCommitLen;\n"
		"		}\n"
		"		{\n"
		"			LangEl *lel = commitHead, *cmStack[numNodes];\n"
		"			int n = absCommitLen, depth = 0, sp = 0;\n"
		"\n"
		"commit_head:\n"
		"			if ( lel->retry > 0 ) {\n"
		"				if ( lel->retry & lower )\n"
		"					numRetry -= 1;\n"
		"				if ( lel->retry & upper )\n"
		"					numRetry -= 1;\n"
		"				lel->retry = 0;\n"
		"			}\n"
		"\n"
		"			/* If depth is > 0 then move over lel freely, otherwise, make\n"
		"			 * sure that we have not already done n steps down the line. */\n"
		"			if ( lel->next != 0 && ( depth > 0 || n > 1 ) ) {\n"
		"				cmStack[sp++] = lel;\n"
		"				lel = lel->next;\n"
		"\n"
		"				/* If we are at the top level count the steps down the line. */\n"
		"				if ( depth == 0 )\n"
		"					n -= 1;\n"
		"				goto commit_head;\n"
		"			}\n"
		"\n"
		"commit_reverse:\n"
		"			if ( lel->child != 0 ) {\n"
		"				cmStack[sp++] = lel;\n"
		"				lel = lel->child;\n"
		"\n"
		"				/* When we move down we need to increment the depth. */\n"
		"				depth += 1;\n"
		"				goto commit_head;\n"
		"			}\n"
		"\n"
		"commit_upwards:\n"
		"			if ( sp > 0 ) {\n"
		"				/* Figure out which place to return to. */\n"
		"				if ( cmStack[sp-1]->next == lel ) {\n"
		"					lel = cmStack[--sp];\n"
		"					goto commit_reverse;\n"
		"				}\n"
		"				else {\n"
		"					/* Going back up, adjust the depth. */\n"
		"					lel = cmStack[--sp];\n"
		"					depth -= 1;\n"
		"					goto commit_upwards;\n"
		"				}\n"
		"			}\n"
		"		}\n"
		"		if ( numRetry == 0 ) {\n"
		"			#ifdef LOG_ACTIONS\n"
		"			cerr << \"number of retries is zero, \"\n"
		"				\"executing final actions\" << endl;\n"
		"			#endif\n"
		"			{\n"
		"				LangEl *lel = commitHead, *cmStack[numNodes];\n"
		"				int sp = 0;\n"
		"				bool doExec = false;\n"
		"\n"
		"final_head:\n"
		"				if ( lel == lastFinal ) {\n"
		"					doExec = true;\n"
		"					goto hit_final;\n"
		"				}\n"
		"\n"
		"				if ( lel->next != 0 ) {\n"
		"					cmStack[sp++] = lel;\n"
		"					lel = lel->next;\n"
		"					goto final_head;\n"
		"				}\n"
		"\n"
		"final_reverse:\n"
		"\n"
		"				if ( lel->child != 0 ) {\n"
		"					cmStack[sp++] = lel;\n"
		"					lel = lel->child;\n"
		"					goto final_head;\n"
		"				}\n"
		"\n"
		"final_upwards:\n"
		"\n"
		"				if ( doExec ) {\n"
		"{\n"
		"	if ( lel->type < " << firstNonTermId << " ) {\n";

	if ( tokenFinalBlock != 0 ) {
		out << "#line " << tokenFinalBlock->loc.line << " \"" << fileName << "\"\n";
		writeInlinList( out, 0, tokenFinalBlock->inlineList );
		out << "\n#line " << outFilter->line + 1 << " \"" << outputFile << "\"\n";
	}

	out <<
		"	}\n"
		"	else {\n"
		"		LangEl *redLel = lel;\n"
		"		if ( redLel->child != 0 ) {\n"
		"			int r = fssProdLengths[redLel->reduction] - 1;\n"
		"			LangEl *rhsEl = redLel->child;\n"
		"			while ( rhsEl != 0 ) {\n"
		"				rhs[r--] = rhsEl;\n"
		"				rhsEl = rhsEl->next;\n"
		"			}\n"
		"		}\n";

	writeFinalBlocks( out );	

	out <<
		"	}\n"
		"}\n"
		"\n"
		"					if ( lel->child != 0 ) {\n"
		"						numNodes -= 1;\n"
		"						LangEl *first = lel->child;\n"
		"						LangEl *child = lel->child;\n"
		"						lel->child = 0;\n"
		"						while ( child->next != 0 ) {\n"
		"							child = child->next;\n"
		"							numNodes -= 1;\n"
		"						}\n"
		"						child->next = pool;\n"
		"						pool = first;\n"
		"					}\n"
		"				}\n"
		"\n"
		"hit_final:\n"
		"				if ( sp > 0 ) {\n"
		"					/* Figure out which place to return to. */\n"
		"					if ( cmStack[sp-1]->next == lel ) {\n"
		"						lel = cmStack[--sp];\n"
		"						goto final_reverse;\n"
		"					}\n"
		"					else {\n"
		"						lel = cmStack[--sp];\n"
		"						goto final_upwards;\n"
		"					}\n"
		"				}\n"
		"\n"
		"				lastFinal = lel;\n"
		"			}\n"
		"		}\n"
		"	}\n"
		"\n"
		"	if ( *action & act_rb ) {\n"
		"		int fssRed = *action >> 2;\n"
		"		int reduction = fssProdIdIndex[fssRed];\n"
		"		LangEl *redLel;\n"
		"		if ( pool == 0 ) {\n"
		"			if ( freshPos == " << FRESH_BLOCK << " ) {\n"
		"				freshEl = new LangEl[" << FRESH_BLOCK << "];\n"
		"				#ifdef LOG_ACTIONS\n"
		"				cerr << \"allocating " << FRESH_BLOCK << " LangEls\" << endl;\n"
		"				#endif\n"
		"				freshPos = 0;\n"
		"			}\n"
		"			redLel = freshEl + freshPos++;\n"
		"		}\n"
		"		else {\n"
		"			redLel = pool;\n"
		"			pool = pool->next;\n"
		"		}\n"
		"		numNodes += 1;\n"
		"		redLel->type = prodLhsIds[reduction];\n"
		"		redLel->reduction = reduction;\n"
		"		redLel->child = 0;\n"
		"		redLel->next = 0;\n"
		"		redLel->retry = (lel->retry << 16);\n"
		"		lel->retry &= upper;\n"
		"\n"
		"		rhsLen = fssProdLengths[fssRed];\n"
		"		if ( rhsLen > 0 ) {\n"
		"			for ( int r = rhsLen-1; r > 0; r-- ) {\n"
		"				rhs[r] = stackTop;\n"
		"				stackTop = stackTop->next;\n"
		"			}\n"
		"			rhs[0] = stackTop;\n"
		"			stackTop = stackTop->next;\n"
		"			rhs[0]->next = 0;\n"
		"		}\n";

	writeReduceBlocks( out );

	out <<
		"		#ifdef LOG_ACTIONS\n"
		"		cerr << \"reduced: \"\n"
		"				<< prodNames[reduction]\n"
		"				<< \" rhsLen: \" << rhsLen;\n"
		"		#endif\n"
		"		if ( action[1] == 0 )\n"
		"			redLel->retry = 0;\n"
		"		else {\n"
		"			redLel->retry += 0x10000;\n"
		"			numRetry += 1;\n"
		"			#ifdef LOG_ACTIONS\n"
		"			cerr << \" retry: \" << redLel;\n"
		"			#endif\n"
		"		}\n"
		"\n"
		"		#ifdef LOG_ACTIONS\n"
		"		cerr << endl;\n"
		"		#endif\n"
		"\n"
		"		if ( rhsLen == 0 ) {\n"
		"			redLel->file = lel->file;\n"
		"			redLel->line = lel->line;\n"
		"			targState = curs;\n"
		"		}\n"
		"		else {\n"
		"			redLel->child = rhs[rhsLen-1];\n"
		"			redLel->file = rhs[0]->file;\n"
		"			redLel->line = rhs[0]->line;\n"
		"			targState = rhs[0]->state;\n"
		"		}\n"
		"\n"
		"		if ( induceReject ) {\n"
		"			#ifdef LOG_ACTIONS\n"
		"			cerr << \"error induced during reduction of \" <<\n"
		"					lelNames[redLel->type] << endl;\n"
		"			#endif\n"
		"			redLel->state = curs;\n"
		"			redLel->next = stackTop;\n"
		"			stackTop = redLel;\n"
		"			curs = targState;\n"
		"			goto parseError;\n"
		"		}\n"
		"		else {\n"
		"			redLel->next = input;\n"
		"			input = redLel;\n"
		"		}\n"
		"	}\n"
		"\n"
		"\n"
		"	curs = targState;\n"
		"	goto again;\n"
		"\n"
		"parseError:\n"
		"	#ifdef LOG_BACKTRACK\n"
		"	cerr << \"hit error\" << endl;\n"
		"	#endif\n";

	if ( undoTransBlock != 0 ) {
		out << "#line " << undoTransBlock->loc.line << " \"" << fileName << "\"\n";
		writeInlinList( out, 0, undoTransBlock->inlineList );
		out << "\n#line " << outFilter->line + 1 << " \"" << outputFile << "\"\n";
	}

	out << 
		"	if ( numRetry > 0 ) {\n"
		"		while ( true ) {\n"
		"			LangEl *redLel = stackTop;\n"
		"			if ( stackTop->type < " << firstNonTermId << " ) {\n"
		"				#ifdef LOG_BACKTRACK\n"
		"				cerr << \"backing up over terminal: \" <<\n"
		"						lelNames[stackTop->type] << endl;\n"
		"				#endif\n"
		"				stackTop = stackTop->next;\n"
		"				redLel->next = input;\n"
		"				input = redLel;\n";

	if ( undoTransBlock != 0 ) {
		out << "#line " << undoTransBlock->loc.line << " \"" << fileName << "\"\n";
		writeInlinList( out, 0, undoTransBlock->inlineList );
		out << "\n#line " << outFilter->line + 1 << " \"" << outputFile << "\"\n";
	}

	out <<
		"			}\n"
		"			else {\n"
		"				#ifdef LOG_BACKTRACK\n"
		"				cerr << \"backing up over non-terminal: \" <<\n"
		"						lelNames[stackTop->type] << endl;\n"
		"				#endif\n"
		"				stackTop = stackTop->next;\n"
		"				LangEl *first = redLel->child;\n"
		"				if ( first == 0 )\n"
		"					rhsLen = 0;\n"
		"				else {\n"
		"					rhsLen = 1;\n"
		"					while ( first->next != 0 ) {\n"
		"						first = first->next;\n"
		"						rhsLen += 1;\n"
		"					}\n"
		"					first->next = stackTop;\n"
		"					stackTop = redLel->child;\n"
		"\n"
		"					LangEl *rhsEl = stackTop;\n"
		"					int p = rhsLen;\n"
		"					while ( p > 0 ) {\n"
		"						rhs[--p] = rhsEl;\n"
		"						rhsEl = rhsEl->next;\n"
		"					}\n"
		"				}\n";

	writeUndoBlocks( out );

	out <<
		"				redLel->next = pool;\n"
		"				pool = redLel;\n"
		"				numNodes -= 1;\n"
		"			}\n"
		"\n"
		"			if ( redLel->retry > 0 ) {\n"
		"				#ifdef LOG_BACKTRACK\n"
		"				cerr << \"found retry targ: \" << redLel << endl;\n"
		"				#endif\n"
		"				numRetry -= 1;\n"
		"				#ifdef LOG_BACKTRACK\n"
		"				cerr << \"found retry: \" << redLel << endl;\n"
		"				#endif\n"
		"				if ( redLel->retry & lower )\n"
		"					curs = input->state;\n"
		"				else {\n"
		"					input->retry = redLel->retry >> 16;\n"
		"					if ( stackTop->state < 0 )\n"
		"						curs = startState;\n"
		"					else {\n"
		"						curs = targs[(int)indicies[offsets[stackTop->state] + \n"
		"							(stackTop->type - keys[stackTop->state<<1])]];\n"
		"					}\n"
		"				}\n"
		"				goto again;\n"
		"			}\n"
		"		}\n"
		"	}\n"
		"	curs = -1;\n"
		"	errCount += 1;\n"
		"	return -1;\n"
		"}\n"
		"\n";

	out << 
		"int " << fsmName << "::finish()\n"
		"{\n"
		"	#ifdef LOG_ACTIONS\n"
		"	cerr << \"numNodes: \" << numNodes << endl;\n"
		"	#endif\n"
		"	return errCount;\n"
		"}"
		"\n";
}

void ParseData::writeTables( ostream &out )
{
	out << "unsigned int startState = " <<
		graph->startState->stateNum << ";\n\n";

	SARRAY_TYPE(maxIndex) << " indicies[] = {\n\t";
	int totalItems = 0;
	for ( StateList::Iter state = graph->stateList; state.lte(); state++ ) {
		for ( TransMap::Iter trans = state->transMap; trans.lte(); trans++ ) {
			out << trans->value->actionSetEl->key.id;
			if ( !state.last() || !trans.last() ) {
				out << ", ";
				if ( ++totalItems % 8 == 0 )
					out << "\n\t";
			}

			if ( ! trans.last() ) {
				TransMap::Iter next = trans.next();
				for ( long key = trans->key+1; key < next->key; key++ ) {
					out << "-1, ";
					if ( ++totalItems % 8 == 0 )
						out << "\n\t";
				}
			}
		}
	}
	out << "\n};\n\n";

	UARRAY_TYPE(maxLelId) << " keys[] = {\n\t";
	totalItems = 0;
	for ( StateList::Iter state = graph->stateList; state.lte(); state++ ) {
		if ( state->transMap.length() == 0 ) {
			out << "0, 0";
			if ( !state.last() ) {
				out << ", ";
				totalItems += 2;
				if ( totalItems % 8 == 0 )
					out << "\n\t";
			}
		}
		else {
			TransMap::Iter first = state->transMap.first();
			TransMap::Iter last = state->transMap.last();
			out << first->key << ", " << last->key;
			if ( !state.last() ) {
				out << ", ";
				totalItems += 2;
				if ( totalItems % 8 == 0 )
					out << "\n\t";
			}
		}
	}
	out << "\n};\n\n";

	UARRAY_TYPE(maxOffset) << " offsets[] = {\n\t";
	totalItems = 0;
	int curOffset = 0;
	for ( StateList::Iter state = graph->stateList; state.lte(); state++ ) {
		out << curOffset;
		if ( !state.last() ) {
			out << ", ";
			if ( ++totalItems % 8 == 0 )
				out << "\n\t";
		}

		/* Increment the offset. */
		if ( state->transMap.length() > 0 ) {
			TransMap::Iter first = state->transMap.first();
			TransMap::Iter last = state->transMap.last();
			curOffset += last->key - first->key + 1;
		}
	}
	out << "\n};\n\n";

	UARRAY_TYPE(maxState) << " targs[] = {\n\t";
	totalItems = 0;
	for ( ActionSet::Iter asi = actionSet; asi.lte(); asi++ ) {
		out << asi->key.targ;
		if ( ! asi.last() ) {
			out << ", ";
			if ( ++totalItems % 8 == 0 )
				out << "\n\t";
		}
	}
	out << "\n};\n\n";

	out << "unsigned int actInds[] = {\n\t";
	totalItems = 0;
	int pos = 0;
	for ( ActionSet::Iter asi = actionSet; asi.lte(); asi++ ) {
		out << pos;
		if ( ! asi.last() ) {
			out << ", ";
			if ( ++totalItems % 8 == 0 )
				out << "\n\t";
		}
		pos += asi->key.actions.length() + 1;
	}
	out << "\n};\n\n";

	out << "unsigned int actions[] = {\n\t";
	totalItems = 0;
	for ( ActionSet::Iter asi = actionSet; asi.lte(); asi++ ) {
		for ( ActDataList::Iter ali = asi->key.actions; ali.lte(); ali++ ) {
			out << *ali << ", ";
			if ( ++totalItems % 8 == 0 )
				out << "\n\t";
		}

		out << "0";
		if ( ! asi.last() ) {
			out << ", ";
			if ( ++totalItems % 8 == 0 )
				out << "\n\t";
		}
	}
	out << "\n};\n\n";

	out << "int commitLen[] = {\n\t";
	totalItems = 0;
	for ( ActionSet::Iter asi = actionSet; asi.lte(); asi++ ) {
		out << asi->key.commitLen;
		if ( ! asi.last() ) {
			out << ", ";
			if ( ++totalItems % 8 == 0 )
				out << "\n\t";
		}
	}
	out << "\n};\n\n";

	out << "unsigned int fssProdIdIndex[] = {\n\t";
	totalItems = 0;
	for ( DefList::Iter prod = prodList; prod.lte(); prod++ ) {
		int numFin = prod->fsm->finStateSet.length();
		for ( int i = 0; i < numFin; i++ ) {
			out << prod->prodId;
			if ( !prod.last() || i < (numFin-1) ) {
				out << ", ";
				if ( ++totalItems % 8 == 0 )
					out << "\n\t";
			}
		}
	}
	out << "\n};\n\n";

	SARRAY_TYPE(maxProdLen) << " fssProdLengths[] = {\n\t";
	totalItems = 0;
	for ( DefList::Iter prod = prodList; prod.lte(); prod++ ) {
		int numFin = prod->fsm->finStateSet.length();
		assert( numFin == 1 );
		for ( int i = 0; i < numFin; i++ ) {
			out << prod->fsmLength;
			if ( !prod.last() || i < (numFin-1)  ) {
				out << ", ";
				if ( ++totalItems % 8 == 0 )
					out << "\n\t";
			}
		}
	}
	out << "\n};\n\n";

	UARRAY_TYPE(maxLelId) << " prodLhsIds[] = {\n\t";
	totalItems = 0;
	for ( DefList::Iter prod = prodList; prod.lte(); prod++ ) {
		out << prod->prodName->id;
		if ( ! prod.last() ) {
			out << ", ";
			if ( ++totalItems % 8 == 0 )
				out << "\n\t";
		}
	}
	out << "\n};\n\n";

	out << "char *prodNames[] = {\n";
	for ( DefList::Iter prod = prodList; prod.lte(); prod++ ) {
		out << "\t\"" << prod->data << "\"";
		if ( ! prod.last() )
			out << ",";
		out << "\n";
	}
	out << "};\n\n";

	out << "char *lelNames[] = {\n";
	for ( int i = 0; i < nextSymbolId; i++ ) {
		KlangEl *lel = langElIndex[i];
		if ( lel != 0 )
			out << "\t\"" << lel->data << "\"";
		else {
			out << "\t\"";
			if ( 33 <= i && i <= 126 ) {
				if ( i == '"' || i == '\\' )
					out << "\\";
				out << (char)i;
			}
			else
				out << "D-" << i;
			out << "\"";
		}
		if ( i < nextSymbolId-1 );
			out << ",";
		out << "\n";
	}
	out << "};\n\n";

	out << "char prodCommit[] = {\n\t";
	totalItems = 0;
	for ( DefList::Iter prod = prodList; prod.lte(); prod++ ) {
		out << ( prod->prodCommit ? '1' : '0' );
		if ( ! prod.last() ) {
			out << ", ";
			if ( ++totalItems % 8 == 0 )
				out << "\n\t";
		}
	}
	out << "\n};\n\n";
}


std::ostream &ParseData::UARRAY_TYPE( unsigned long long maxVal )
{
	ostream &out = *outStream;
	if ( maxVal <= UCHAR_MAX )
		out << "unsigned char";
	else if ( maxVal <= USHRT_MAX )
		out << "unsigned short";
	else if ( maxVal <= UINT_MAX )
		out << "unsigned int";
	else if ( maxVal <= ULONG_MAX )
		out << "unsigned long";
	else
		out << "unsigned long long";
	return out;
}

std::ostream &ParseData::SARRAY_TYPE( signed long long maxVal )
{
	ostream &out = *outStream;
	if ( maxVal <= CHAR_MAX )
		out << "char";
	else if ( maxVal <= SHRT_MAX )
		out << "short";
	else if ( maxVal <= INT_MAX )
		out << "int";
	else if ( maxVal <= LONG_MAX )
		out << "long";
	else 
		out << "long long";
	return out;
}
